# -*- coding: utf-8 -*-

import pandas

from base_parser_class import InteropBinParser

class InteropErrorMetrics( InteropBinParser ):

    __version = 0.1
    supported_versions = [3]
    codename = 'error'

    def _init_variables(self):
        self.data = { 'lane': [], 
                      'tile': [], 
                      'cycle': [], 
                      'perfect': [], 
                      'rate': [],
                      'one_err': [], 
                      'two_err': [], 
                      'three_err': [], 
                      'four_err': [] }
            
    def parse_binary(self):
    
        bs = self.bs
    
        # Contains cycle error rate as well as counts for perfect reads and read with 1-4 errors
        # Format:
        #   byte 0: file version number (3)
        #   byte 1: length of each record
        #   bytes (N * 30 + 2) to (N * 30 + 11): record:
        #       2 bytes: lane number (uint16)
        #       2 bytes: tile number (uint16)
        #       2 bytes: cycle number (uint16)
        #	4 bytes: error rate (float)
        #	4 bytes: number of perfect reads (uint32)
        #	4 bytes: number of reads with 1 error (uint32)
        #	4 bytes: number of reads with 2 errors (uint32)
        # 	4 bytes: number of reads with 3 errors (uint32)
        #	4 bytes: number of reads with 4 errors (uint32)
        #   ...where N is the record index.

        self.apparent_file_version = bs.read('uintle:8')
        
        self.check_version(self.apparent_file_version)
        
        recordlen = bs.read('uintle:8')  # length of each record

        for i in range(0,((bs.len) / (recordlen * 8))):  # record length in bits
            self.data['lane'].append(bs.read('uintle:16'))
            self.data['tile'].append(bs.read('uintle:16'))
            self.data['cycle'].append(bs.read('uintle:16'))
            self.data['rate'].append(bs.read('floatle:32'))
            self.data['perfect'].append(bs.read('uintle:32'))
            self.data['one_err'].append(bs.read('uintle:32'))
            self.data['two_err'].append(bs.read('uintle:32'))
            self.data['three_err'].append(bs.read('uintle:32'))
            self.data['four_err'].append(bs.read('uintle:32'))

        self.df = pandas.DataFrame(self.data)
    
    def __str__(self):
        #TODO: to_str (improve output)
        out = "(sum of all types of errors across all reads)\n"
        idf = self.make_coordinate_plane(self.df)
        out += "%s\n" % idf.sum()
        return out

    
if __name__=='__main__':
    
    import sys
    
    try:
        filename = sys.argv[1]
    except:
        print 'supply path to ErrorMetrics.bin (or ErrorMetricsOut.bin)'
        sys.exit()
    
    EM = InteropErrorMetrics(filename)
    
    print 'Length of data: %i' % len(EM.data['cycle'])
    print EM
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# 3/28/2013
# by nthmost (naomi.most@invitae.com)

import pandas
from bitstring import ReadError

from base_parser_class import InteropBinParser

class InteropControlMetrics(InteropBinParser):

    __version = 0.1
    supported_versions = [1]
    codename = 'control'

    def _init_variables(self):
        self.data = { 'lane': [], 
                      'tile': [], 
                      'read': [], 
                      'control_str': [], 
                      'index_str': [], 
                      'clusters': [] 
                    }

    def parse_binary(self):
    
        bs = self.bs

        # Control Metrics (ControlMetricsOut.bin)
        # Contains pull out information for Illumina in-line sample controls
        # Format:
        #   byte 0: file version number (1) bytes (variable length): record:
        #   2 bytes: lane number (uint16)
        #   2 bytes: tile number (uint16)
        #   2 bytes: read number (uint16)
        #   2 bytes: number bytes X for control name(uint16)
        #   X bytes: control name string (string in UTF8Encoding) 
        #   2 bytes: number bytes Y for index name(uint16)
        #   Y bytes: index name string (string in UTF8Encoding) 
        #   4 bytes: num of clusters identified as control (uint32)

        self.apparent_file_version = bs.read('uintle:8')  # version number of binary 
        self.check_version(self.apparent_file_version)

        try:
            while True:
                self.data['lane'].append(bs.read('uintle:16'))
                self.data['tile'].append(bs.read('uintle:16'))
                self.data['read'].append(bs.read('uintle:16'))
    
                # next 2 bytes: expected control name length in bytes.
                nextbytes = bs.read('uintle:16')
                self.data['control_str'].append(bs.read('bytes:%i' % (nextbytes)))

                # next 2 bytes: expected index name length in bytes.    
                nextbytes = bs.read('uintle:16')
                self.data['index_str'].append(bs.read('bytes:%i' % (nextbytes)))
            
                self.data['clusters'].append(bs.read('uintle:32'))
                
        except ReadError:
            pass
    
        self.df = pandas.DataFrame(self.data)

    def __str__(self):
        #TODO: to_str (improve output)
        out = "%s\n" % self.df.head()
        return out


if __name__=='__main__':
    import sys
    
    try:
        filename = sys.argv[1]
    except:
        print "supply path to ExtractionMetrics.bin"
        sys.exit()
    
    CM = InteropControlMetrics(filename)
    print CM



# -*- coding: utf-8 -*-

import pandas
from bitstring import BitString

from base_parser_class import InteropBinParser

class InteropCorrectedIntensityMetrics(InteropBinParser):

    __version = 0.1
    supported_versions = [2]
    codename = 'corint'

    def _init_variables(self):
        self.data = { 
                 'lane': [], 
                 'cycle': [], 
                 'tile': [], 
                 'avg_intensity': [], 
                 'avg_corint_A': [], 
                 'avg_corint_C': [], 
                 'avg_corint_T': [], 
                 'avg_corint_G': [],
                 'avg_corint_called_A': [], 
                 'avg_corint_called_C': [], 
                 'avg_corint_called_G': [], 
                 'avg_corint_called_T': [],
                 'num_nocalls': [], 
                 'num_calls_A': [], 
                 'num_calls_C': [], 
                 'num_calls_G': [], 
                 'num_calls_T': [],
                 'signoise_ratio': [] 
               }

    def parse_binary(self):
    
        bs = self.bs    #convenience assignment

        self.apparent_file_version = bs.read('uintle:8')
        self.check_version( self.apparent_file_version )
        
        recordlen = bs.read('uintle:8')  # length of each record

        # CorrectedIntMetrics.bin / CorrectedIntMetricsOut.bin 
        #
        # Contains base call metrics
        # Format:
        #   byte 0: file version number (2)
        #   byte 1: length of each record
        #   bytes (N * 48 + 2) to (N * 48 + 49): record:
        #       2 bytes: lane number (uint16)
        #       2 bytes: tile number (uint16)
        #       2 bytes: cycle number (uint16)
        #	2 bytes: average intensity (uint16) 
        #	2 bytes: average corrected int for channel A (uint16)
        #	2 bytes: average corrected int for channel C (uint16)
        #	2 bytes: average corrected int for channel G (uint16)
        #	2 bytes: average corrected int for channel T (uint16)
        #	2 bytes: average corrected int for called clusters for base A (uint16)
        #	2 bytes: average corrected int for called clusters for base C (uint16)
        #	2 bytes: average corrected int for called clusters for base G (uint16)
        #	2 bytes: average corrected int for called clusters for base T (uint16)
        #	20 bytes: number of base calls (float) for No Call and channel [A, C, G, T] respectively
        #	4 bytes: signal to noise ratio (float)

        for i in range(0,((bs.len) / (recordlen * 8 ))):  # record length in bits
            self.data['lane'].append(bs.read('uintle:16'))
            self.data['tile'].append(bs.read('uintle:16'))
            self.data['cycle'].append(bs.read('uintle:16'))
            self.data['avg_intensity'].append(bs.read('uintle:16'))
            self.data['avg_corint_A'].append(bs.read('uintle:16'))
            self.data['avg_corint_C'].append(bs.read('uintle:16'))
            self.data['avg_corint_G'].append(bs.read('uintle:16'))
            self.data['avg_corint_T'].append(bs.read('uintle:16'))
            self.data['avg_corint_called_A'].append(bs.read('uintle:16'))
            self.data['avg_corint_called_C'].append(bs.read('uintle:16'))
            self.data['avg_corint_called_G'].append(bs.read('uintle:16'))
            self.data['avg_corint_called_T'].append(bs.read('uintle:16'))
    
            # 20 bytes / 5 = 4 bytes each following records.
            self.data['num_nocalls'].append(bs.read('floatle:32'))
            self.data['num_calls_A'].append(bs.read('floatle:32'))
            self.data['num_calls_C'].append(bs.read('floatle:32'))
            self.data['num_calls_G'].append(bs.read('floatle:32'))
            self.data['num_calls_T'].append(bs.read('floatle:32'))
    
            # 4 bytes: sig/noise ratio (float)  
            self.data['signoise_ratio'].append(bs.read('floatle:32'))


        self.df = pandas.DataFrame(self.data)

        # place each metric into a coordinate plane so we can sort into reads.
        self.idf = self.make_coordinate_plane(self.df)

    def __str__(self):
        #TODO: to_str (improve output)
        out = "%i entries in CorrectedIntensityMetrics binary" % len(self.data['cycle'])
        out = "\nSample from lane/cycle/tile start:"
        out += "%s\n" % self.idf.head()
        return out
 


if __name__=='__main__':

    import sys
    
    try:
        filename = sys.argv[1]
    except:
        print "supply path to CorrectedIntensityMetrics.bin"
        sys.exit()

    CIM = InteropCorrectedIntensityMetrics(filename)
    print CIM



# -*- coding: utf-8 -*-

import cStringIO
from bitstring import BitString

from utils import dmesg

#### SEQUENCER VAGARIES: flowcell_layout and read_config
#
# All binary parsers use these dicts, though each parser can have a different 
# flowcell layout set either by supplying these dicts as keyword arguments to class instantiation, 
# or by explicitly setting the flowcell_layout or read_config attrib after instantiation.
#
# Typical MiSeq parameters would be as follows:

FLOWCELL_LAYOUT_DEFAULTS = { 'lanecount': 1, 'surfacecount': 2, 'swathcount': 1, 'tilecount': 14 }

# Typical Read Configuration from a MiSeq:

READ_CONFIG_DEFAULTS = [{'read_num': 1, 'cycles': 151, 'is_index': 0}, 
                        {'read_num': 2, 'cycles': 6, 'is_index': 1}, 
                        {'read_num': 3, 'cycles': 151, 'is_index': 0}]

class InteropBinParser(object):
    "Generic binary parser for ILMN files typically found in InterOp directory. Subclass (do not use directly)."

    __version = 0.5      # version of this base class
    
    num_tiles = 0
    num_reads = 0

    def __init__(self, bitstring_or_filename, **kwargs):
        "Takes either a filename or a BitString object. Optional: flowcell_layout {}, read_config [{},]"

        self.flowcell_layout = kwargs.get('flowcell_layout', FLOWCELL_LAYOUT_DEFAULTS)
        self.read_config = kwargs.get('read_config', READ_CONFIG_DEFAULTS)

        # see if it's a filename or a bitstring (aka bitstream)
        try:
            bitstring_or_filename.all(1)    # attempts to perform the "are these bits all 1s" method
            self.bs = bitstring_or_filename
        except AttributeError:              # assume it's a filename, then.
            self.bs = BitString(bytes=open(bitstring_or_filename, 'rb').read())
            
        self.num_tiles = reduce(lambda x, y: x*y, self.flowcell_layout.values())
        self.num_reads = len(self.read_config)

        self._init_variables()

        if self.bs is None:
            dmesg("[%s] Fatal: BitString could not be created (file empty or not found)." % 
                    (self.__class__.__name__), 1)
        else:
            self.parse_binary() 

    def parse_binary(self):
        "Stub class for binary parsing."
        print "InteropBinParser: Generic Binary Parser class"
        print ""
        print "Override this method with your own parsing method." 
        print ""

    def _init_variables(self):
        "Place to initialize the instance variables required by specific parsers."
        pass
    
    def check_version(self, version_num):
        "Compare parsed binary's version against parser's supported_versions list."
        
        if version_num not in self.supported_versions:        
            dmesg("[%s] Warning: apparent file version (%i) may not be supported by this parser" % 
                (self.__class__.__name__, self.apparent_file_version), 2)

    def make_coordinate_plane(self, df, flatten=False):
        """Rework a dataframe containing lane / tile / cycle columns into a new dataframe using 
           lane-tile-cycle as a combined index -- sort of a coordinate plane."""
           
        if flatten:
            df = df.copy()
            # recast the coordinate system as a descriptive index composed like so:
            # cycle * 1000000 + lane * 10000 + tile
            
            df.lane = df.lane.mul(10000)
            df.cycle = df.cycle.mul(1000000)

            # ...that way the index stays human-readable and still easily sorted and sliced.

        idf = df.set_index(['cycle','lane','tile'])

        if flatten:
            idf.index = idf.index.map(sum)
        
        return idf.sort()

    def to_dict(self):
        "Parser subclasses should override this, make it more specifically relevant."
        return self.data

    def to_csv(self):
        "Transforms object's DataFrame into comma-separated / newline delineated data. First line contains row headers." 
        output = cStringIO.StringIO()
        self.df.to_csv(output, index_label='line')
        return output.getvalue()

    def to_json(self):
        "Transforms object's DataFrame into a json document."
        return self.df.to_json()





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# InteropDataset
# Library encapsulating the XML and bin files from MiSeq and HiSeq output.
#
# InteropMetadata
# Parser for XML files from MiSeq / HiSeq run data.
#
# See README for intro and basic examples.
#
# March 2013
# by nthmost (naomi.most@invitae.com)
# with lots of help from ECO (eric.olivares@invitae.com)

import time, os

import pandas

from metadata import InteropMetadata
from index_metrics import InteropIndexMetrics
from tile_metrics import InteropTileMetrics
from quality_metrics import InteropQualityMetrics
from error_metrics import InteropErrorMetrics
from corint_metrics import InteropCorrectedIntensityMetrics 
from control_metrics import InteropControlMetrics
from extraction_metrics import InteropExtractionMetrics

from utils import select_file_from_aliases
from exceptions import InteropFileNotFoundError
from filemaps import BINFILE_DIR_NAME, XML_FILEMAP, BIN_FILEMAP


class InteropDataset(object):
    """Encapsulates the physical files related to this sequencing run. 
       Absolves other classes of having to know about files and directories.
       Combines parser results to emulate Illumina Sequencing Analysis Viewer."""

    directory = ""  # supplied directory name to instantiate class (relative path)
    fullpath = ""   # calculated absolute filesystem path after instantiation
    xmldir = ""     # typically equivalent to fullpath
    bindir = ""     # typically = fullpath/Interop/

    meta = None

    def __init__(self, targetdir):
        "Supply a path (directory) that should contain XML files, with an InterOp directory within it."

        self.directory = targetdir

        self.xmldir = self.directory
        self.bindir = os.path.join(self.directory, BINFILE_DIR_NAME)
        
        # aggregate results of parsing one or more XML files        
        self.meta = self.Metadata()
    
        # holders for the parser objects. Reference via *Metrics() classes.
        self._quality_metrics = None
        self._tile_metrics = None
        self._index_metrics = None
        self._error_metrics = None
        self._corint_metrics = None
        self._extraction_metrics = None
        self._control_metrics = None

    def get_binary_path(self, codename):
        "returns absolute path to binary file represented by data 'codename'"
        path = select_file_from_aliases(codename, BIN_FILEMAP, self.bindir)
        if path is None:
            raise InteropFileNotFoundError("No suitable binary found for {} in directory {}".format(codename, self.bindir))
        else:
            return path
    
    def Metadata(self, reload=False):
        "returns InteropMetadata class generated from this dataset's XML files"
        if self.meta == None or reload == True:
            self.meta = InteropMetadata(self.xmldir)
        return self.meta
    
    def QualityMetrics(self, reload=False):
        "Returns InteropQualityMetrics object from the 'quality' binary in this dataset."
        if self._quality_metrics == None or reload == True:
            self._quality_metrics = InteropQualityMetrics(self.get_binary_path('quality'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._quality_metrics
        
    def TileMetrics(self, reload=False):
        "Returns InteropTileMetrics object from the 'tile' binary in this dataset."
        if self._tile_metrics == None or reload == True:
            self._tile_metrics = InteropTileMetrics(self.get_binary_path('tile'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._tile_metrics

    def IndexMetrics(self, reload=False):
        "Returns InteropIndexMetrics object from the 'index' binary in this dataset."
        if self._index_metrics == None or reload == True:
            self._index_metrics = InteropIndexMetrics(self.get_binary_path('index'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._index_metrics

    def ControlMetrics(self, reload=False):
        "Returns InteropControlMetrics object from the 'control' binary in this dataset."
        if self._control_metrics == None or reload == True:
            self._control_metrics = InteropControlMetrics(self.get_binary_path('control'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._control_metrics

    def ErrorMetrics(self, reload=False):
        "Returns InteropErrorMetrics object from the 'error' binary in this dataset."
        if self._error_metrics == None or reload == True:
            self._error_metrics = InteropErrorMetrics(self.get_binary_path('error'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._error_metrics

    def ExtractionMetrics(self, reload=False):
        "Returns InteropExtractionMetrics object from the 'extraction' binary in this dataset."
        if self._extraction_metrics == None or reload == True:
            self._extraction_metrics = InteropExtractionMetrics(self.get_binary_path('extraction'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._extraction_metrics

    def CorrectedIntensityMetrics(self, reload=False):
        "Returns InteropCorrectedIntensityMetrics object from the 'corint' binary in this dataset."
        if self._corint_metrics == None or reload == True:
            self._corint_metrics = InteropCorrectedIntensityMetrics(self.get_binary_path('corint'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._corint_metrics

#TODO: ImageMetrics
#    def ImageMetrics(self, reload=False):
#        "returns InteropImageMetrics object from the 'image' binary in this dataset"
#        if self._image_metrics == None or reload == True:
#            self._image_metrics = InteropImageMetrics(self.get_binary_path('image'), 
#                                    flowcell_layout=self.meta.flowcell_layout,
#                                    read_config=self.meta.read_config )
#        return self._image_metrics
    
  
## Command Line helper functions below
def print_sample_dataset(ID):
    meta = ID.Metadata()

    print ""
    print "METADATA"
    print "--------"
    print meta
    print ""
    
    print "Resequencing Statistics:" 
    print meta.resequencing_stats
    print ""
    
    tm = ID.TileMetrics()
    print "SUMMARY"
    print "-------"
    print "(tile metrics)\n"
    print tm
    print ""
    
    # QualityMetrics usually take a while...
    print "QUALITY"
    print "-------"
    qm = ID.QualityMetrics()
    print "(% >= Q30 per read)\n"
    print qm
    print ""

    im = ID.IndexMetrics()
    print "INDEXING"
    print "--------\n"
    print "Total Reads: %i" % tm.num_clusters
    print "Reads PF: %i" % tm.num_clusters_pf
    print "Percentage Reads Identified (PF): %f" % (float(im.total_ix_reads_pf / tm.num_clusters_pf)*100)
    print ""
    print im
    print ""

    print "ERRORS"
    print "------"
    try:
        em = ID.ErrorMetrics()
        print "(sum of all types of errors across all reads)"
        idf = em.make_coordinate_plane(em.df)
        print idf.sum()
    except InteropFileNotFoundError:
        print "None. (no error metrics binary in this dataset.)"
    finally:
        print ""

    """
    print "INTENSITY"
    print "---------"
    try:
        cm = ID.CorrectedIntensityMetrics()
        print "(sample of raw data)"
        if cm is not None:
            print cm.idf.head()
        else:
            print "CorrectedIntensityMetrics parsing failed."
    except TypeError:
        print "No CorrectedIntensityMetrics binary in this dataset."
    finally:
        print ""
    """

if __name__=='__main__':
    import sys

    try:
        dirname = sys.argv[1] 
    except IndexError:
        print "Supply the absolute or relative path to a directory of sequencing rundata."
        print "Example:  python interop.py /home/username/seqruns/2013-02/0"
        sys.exit()
        
    myDataset = InteropDataset(dirname)

    #print myDataset.meta.read_config

    print_sample_dataset(myDataset)
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# InteropDataset
# Library encapsulating the XML and bin files from MiSeq and HiSeq output.
#
# InteropMetadata
# Parser for XML files from MiSeq / HiSeq run data.
#
# See README for intro and basic examples.
#
# March 2013
# by nthmost (naomi.most@invitae.com)
# with lots of help from ECO (eric.olivares@invitae.com)

import time, os
from collections import OrderedDict
from xml.etree import ElementTree as ET

from exceptions import InteropFileNotFoundError
from utils import select_file_from_aliases
from filemaps import XML_FILEMAP

class InteropMetadata(object):
    """Parser for sequencer's XML files describing a single run. Supply with directory to instantiate.

    CHANGES:
        
        0.2     cleaner logical process for using the various XML files. No longer throws exceptions.
        0.1     first released version.
    """
    
    __version = 0.2     # version of this parser.

    def __init__(self, xmldir):
        """Takes the absolute path of a sequencing run data directory as sole required variable.
           Attempts to parse CompletedJobInfo.xml (or viable alias). If not available, uses 
           runParameters.xml and/or runInfo.xml, which have some overlapping info (but not all).
           
           Individual parsers can be explicitly called via their respective methods.
           
           Be aware that parsing methods are DESTRUCTIVE to existing instance data."""

        self.xmldir = xmldir
        self.experiment_name = ""        # "RU1453:::/locus/data/run_data//1337/1453"
        self.investigator_name = ""      # "Locus:::Uncle_Jesse - 612 - MiSeq"
        self.runID = ""                  # cf CompletedJobInfo.xml / RTARunInfo / Run param "Id"
        self.start_datetime = None
        self.end_datetime = None
        self.rta_run_info = { }

        # read_config: a list of dictionaries, each of which describe a single read from the sequencer.
        self.read_config = []
        
        # Flow cell layout: necessary to enable parsing of different machine types' binaries.
        self.flowcell_layout = { }
                
        # Read numbers from ResequencingRunStats.xml 
        # Example: { 'clusters_raw': 19494893, 'clusters_PF': 17381252, 
        #            'unindexed': 508055, 'unindexed_PF': 16873197, 
        #            'unaligned': 18572490, 'unaligned_PF': 16973197 }   
        self.resequencing_stats = {}

        if self.get_xml_path('reseqstats') is not None:
            self.parse_ResequencingRunStats(self.get_xml_path('reseqstats'))
        
        # CompletedJobInfo.xml is the most complete XML file in the set for what we need,
        # However, only RunInfo.xml and/or RunParameters.xml will be available during an active run.
        # Getting read_config and flowcell_layout are top priority since they inform the binary parsers.
        #
        # TODO: xml_flex (proposed improvement allowing a config file to set which tokens are required / not required.)
        #       Also we might want to specify priority of provenance (e.g. get start_datetime from 'runparams' first).
        #       If you (yes YOU) have any opinions about this, please email me: naomi.most@invitae.com

        self._xml_map = OrderedDict({ 'completed': [None, self.parse_CompletedJobInfo], 
                                      'runinfo':   [None, self.parse_RunInfo],
                                      'runparams': [None, self.parse_RunParameters] })
        self._set_xml_map()
        
        # cycle through XML files, filling from what's available.
        for codename in self._xml_map:
            if self._xml_map[codename][0] is not None:
                self._xml_map[codename][1](self._xml_map[codename][0])
                if codename == 'completed':
                    break

    def _set_xml_map(self):
        "finds all available XML files and assigns them to an ordered dictionary mapping of codename:[filepath,parse_function]"
        for codename in self._xml_map:
            self._xml_map[codename][0] = self.get_xml_path(codename)
                
    def get_xml_path(self, codename):
        "returns absolute path to XML file represented by data 'codename' or None if not available."
        result = select_file_from_aliases(codename, XML_FILEMAP, self.xmldir)
        return result

    def parse_Run_ET(self, run_ET):
        "parses chunk of XML associated with the RTA Run Info blocks in (at least) 2 xml files."

        self.rta_run_info = { 'instrument': run_ET.find("Instrument").text,     # M00612
                              'flowcell': run_ET.find("Flowcell").text,         # 000000000-A316T
                              'date': run_ET.find("Date").text }                # 130208
    
        flowcell_ET = run_ET.find("FlowcellLayout")
        self.flowcell_layout = { 'lanecount': int(flowcell_ET.attrib['LaneCount']),
                        'surfacecount': int(flowcell_ET.attrib['SurfaceCount']),
                        'swathcount': int(flowcell_ET.attrib['SwathCount']),
                        'tilecount': int(flowcell_ET.attrib['TileCount']) }     
        
        # Run / Reads - describes number of cycles per read (and if read is an Index) in sequencing run.
        
        # Because parsing is understood to be destructive, and Reads can be found in multiple files,
        # we start by emptying out whatever's currently in the read_config array for this instance.
        
        self.read_config = []
        read_num = 0
        for item in run_ET.find("Reads"):
            read_num += 1       # redundant safety assignment to read_num
            self.read_config.append( {'read_num': read_num,
                                      'cycles': int(item.attrib['NumCycles']), 
                                      'is_index': True if item.attrib["IsIndexedRead"]=="Y" else False } )      
    
    def parse_ResequencingRunStats(self, filepath):
        """Parses ResequencingRunStatistics.xml (or viable alias) to fill instance variables."""
        
        tree = ET.parse(filepath)
        root = tree.getroot()   # should be "StatisticsResequencing"
        runstats_ET = root.find("RunStats")
        
        self.resequencing_stats = { 'clusters_raw': int(runstats_ET.find("NumberOfClustersRaw").text),
                                    'clusters_pf': int(runstats_ET.find("NumberOfClustersPF").text),
                                    'unindexed': int(runstats_ET.find("NumberOfUnindexedClusters").text),
                                    'unindexed_pf': int(runstats_ET.find("NumberOfUnindexedClustersPF").text),
                                    'unaligned': int(runstats_ET.find("NumberOfUnalignedClusters").text),
                                    'unaligned_pf': int(runstats_ET.find("NumberOfUnalignedClustersPF").text),
                                    'duplicate': int(runstats_ET.find("NumberOfDuplicateClusters").text) }

    def parse_CompletedJobInfo(self, filepath):
        """parses CompletedJobInfo.xml (or viable alias) to fill instance variables."""

        # comments show example data from a real MiSeq run (2013/02)
        
        tree = ET.parse(filepath)
        root = tree.getroot()       #should be "AnalysisJobInfo"
    
        # Something to be aware of: RTARunInfo contains a "version" attribute.
        # (This parser knows how to deal with version 2.)
        self.rta_version = root.find("RTARunInfo").attrib['Version']

        # original location of data output from the sequencer.
        self.output_folder = root.find("RTAOutputFolder").text 
        
        # TODO: reformat these dates into something nicer (more useable).
        self.start_datetime = root.find("StartTime").text           # 2013-02-09T15:51:50.0811937-08:00
        self.end_datetime = root.find("CompletionTime").text        # 2013-02-09T16:06:44.0124452-08:00
        
        # dechunk all of the major sections we want to extract data from.
        sheet_ET = root.find("Sheet")
        header_ET = sheet_ET.find("Header")
        run_ET = root.find("RTARunInfo").find("Run")
        
        # Sheet / *
        try:
            self.runtype = sheet_ET.find("Type").text       # MiSeq, HiSeq, etc.
        except AttributeError:
            #older (early 2012) XML files have no "Type" token.
            self.runtype = ""
                
        # Sheet / Header / *
        # TODO: xml_flex
        try:
            self.investigator_name = header_ET.find("InvestigatorName").text
            self.project_name = header_ET.find("ProjectName").text
            self.experiment_name = header_ET.find("ExperimentName").text
        except AttributeError:
            pass

        # RTARunInfo / Run / *          
        self.runID = run_ET.attrib["Id"] 
        self.parse_Run_ET(run_ET)

    def parse_RunParameters(self, filepath):
        """partially implemented, not essential. Can fill read_config but not flowcell_layout."""

        tree = ET.parse(filepath)
        root = tree.getroot()

        # TODO: xml_flex - designate source priority for tokens found in multiple sources. 
        #self.runID = root.find("RunID").text
        
        # TODO: normalize this variable
        #self.start_datetime = root.find('RunStartDate').text    # format: 130208 YYMMDD
        try:
            self.experiment_name = root.find('ExperimentName').text
        except AttributeError:
            pass

        self.read_config = []
        for item in root.find('Reads'):
            #Different format from that in CompletedJobInfo.xml (contains read Number)
            self.read_config.append( {'read_num': int(item.attrib['Number']),
                                      'cycles': int(item.attrib['NumCycles']), 
                                      'is_index': True if item.attrib['IsIndexedRead']=='Y' else False } )


    def parse_RunInfo(self, filepath):
        "parses Reads, Date, Flowcell, Instrument out of runInfo.xml"
        tree = ET.parse(filepath)
        run_ET = tree.getroot().find('Run')     #little of use in this file except <Run> subelement. 
        
        self.runID = run_ET.attrib['Id']
        
        #? is this useful information? if so, what for?  (It's not in CompletedJobInfo.xml)
        #self.runNumber = run_ET.attrib['Number']    
        
        self.rta_run_info = self.parse_Run_ET(run_ET)
        
    def prettyprint_read_config(self):
        out = "Read Config:"
        for read in self.read_config:
            out += "    Read %i: %i cycles %s" % (read['read_num'], read['cycles'], 
                                    "(Index)" if read['is_index'] else "")
        return out

    def prettyprint_flowcell_layout(self):
        out = """Flowcell Layout:
        Tiles: %(tilecount)i
        Lanes: %(lanecount)i
        Surfaces: %(surfacecount)i
        Swaths: %(swathcount)i""" % self.flowcell_layout
        return out

    def __str__(self):
        "Print the most important metadata (flowcell layout and read config)"
        out = self.prettyprint_read_config() + "\n"
        out += self.prettyprint_flowcell_layout() + "\n"
        return out





#### CONFIGURABLE THINGS
#
# THIS STUFF WILL BE MOVED INTO A MORE PYTHONIC CONFIG INI TYPE THING.
#
# Configuration of filenames, paths, and types/amounts of output logging.
# (Most people will not need to edit this stuff.)

# Name of directory containing Interop binaries. Usually "InterOp".
BINFILE_DIR_NAME = "InterOp"

# BINARY and XML FILEMAPs
#
# "codenames" (rather than filenames) are used internally in InteropDataset to refer to files.
# (The files themselves contain no explicit indication of what's inside them.)
#
# The FILEMAP variables contain mappings of codename to filename, using aliases.
# to select file, since there are several different filenames out there.
# 
# The first filename to be retrievable will be the one that gets parsed, so view the list as
# a set of decreasing priorities.

BIN_FILEMAP = { 'extraction': ["ExtractionMetricsOut.bin", "ExtractionMetrics.bin"],
                'quality': ["QMetricsOut.bin", "QualityMetricsOut.bin", "QualityMetrics.bin"],
                'error': ["ErrorMetricsOut.bin", "ErrorMetrics.bin"],
                'tile': ["TileMetricsOut.bin", "TileMetrics.bin"],
                'corint': ["CorrectedIntMetricsOut.bin", "CorrectedIntensityMetricsOut.bin", "CorrectedIntMetrics.bin"],
                'control': ["ControlMetricsOut.bin", "ControlMetrics.bin"],
                'image': ["ImageMetricsOut.bin", "ImageMetrics.bin"], 
                'index': ["IndexMetricsOut.bin", "IndexMetrics.bin"] }

XML_FILEMAP = { 'runinfo': ["RunInfo.xml"],
               'runparams': ["runParameters.xml"],
               'reseqstats': ["ResequencingRunStatistics.xml"],
               'completed': ["CompletedJobInfo.xml"] }

#### END OF CONFIGURABLE THINGS ####




import time, os

def select_file_from_aliases(codename, filemap, basepath):
    "acquires first physical file available in dataset matching codename."
    #assume that list in filemap has first item as most likely file.
    for item in filemap[codename]:
        filepath = os.path.join(basepath, item)
        if os.path.exists(filepath):
            return filepath
    # none of the possible files extant
    return None

def dmesg(msg, lvl=1):
    VERBOSITY = 1
    DEBUG = True
    "Simple logger - transitional code."
    msg = "[%s] " % str(time.time()) + str(msg)
    if DEBUG: msg = "[DEBUG] " + msg
    if VERBOSITY >= lvl:
        print msg
    
def set_column_sequence(dataframe, seq):
    '''
    Returns pandas DataFrame with the columns named in seq as first columns.
    
    :param dataframe: pandas DataFrame to be resorted according to seq.
    :param seq: array of column names.
    '''
    # http://stackoverflow.com/questions/12329853/how-to-rearrange-pandas-column-sequence
    cols = seq[:] # copy so we don't mutate seq
    for x in dataframe.columns:
        if x not in cols:
            cols.append(x)
    return dataframe[cols]
    


from interop import InteropDataset, print_sample_dataset 
from exceptions import InteropFileNotFoundError
from metadata import InteropMetadata
from base_parser_class import InteropBinParser
from tile_metrics import InteropTileMetrics
from quality_metrics import InteropQualityMetrics
from index_metrics import InteropIndexMetrics
from error_metrics import InteropErrorMetrics
from control_metrics import InteropControlMetrics
from corint_metrics import InteropCorrectedIntensityMetrics
from extraction_metrics import InteropExtractionMetrics



from __future__ import print_function

import os, sys, time

from docopt import docopt
from .interop import InteropDataset 
from .exceptions import InteropFileNotFoundError

__doc__="""ILLUMINATE

Usage: illuminate [options] <datapath>
       illuminate [options] [--csv | --json] <datapath>

By default, illuminate prints a summary of most commonly desired characteristics rather
than raw data (e.g. cluster density from --tile, Q30 percentage scores from --quality.)

Raw data can be output to --csv or --json, either to STDOUT or to file(s). If no --outpath
specified, data will be sent to STDOUT with two newlines separating each metric section.

The --outpath / -o param should contain an already existing directory which the user has
permissions to create new directories within.

For example, 

  illuminate --csv --quality --extraction -o /data/dump/ /path/to/dataset 

produces:  /data/dump/name/quality.metrics.csv, /data/dump/name/extraction.metrics.csv

...where `name` is either a user-supplied --name parameter or the RunID given by the 
sequencer (as recorded in RTA_Run_Info).

This utility is undergoing rapid development; please treat as Very Beta. --NM 2/21/2014

  -h --help             Show this screen.
  --version             Show version.
  -v, --verbose         Increase verbosity           
  -q, --quiet           Suppress all console output   
  -d, --debug           Increase verbosity and prefix output with Unix timestamps. 
  -i, --interactive     Load dataset into iPython for interactive fun.
  -n --name=name        Set a name for this dataset. [default: meta.runID]
  
  --all             Parse and print (or dump) everything
  --meta            Print flowcell_layout and read_config

  --tile            Parse tile metrics
  --quality         Parse quality metrics 
  --index           Parse index metrics
  --error           Parse error metrics
  --corint          Parse corrected intensity metrics
  --extraction      Parse extraction metrics
  --control         Parse control metrics

  --csv             Output raw data from parser as CSV 
  --json            Output raw data from parser as JSON
  
  -o, --outpath=<outpath> Output parser results to directory
  -t, --timestamp   Generate filename(s) containing Unix timestamps (format: timestamp.metric.format)
"""

#TODO: SAV_emu
"""
  --analysis        Produce an emulation of the Analysis screen
  --summary         Produce an emulation of the Summary screen
  --indexing        Produce an emulation of the Indexing screen
  --image           Parse image metrics (no plans to support)
"""

VERBOSITY = 1
DEBUG = False

DEFAULT_FNAME = '%s.%s'  #codename, suffix
TIMESTAMP_FNAME = '%i.%s.%s'   # timestamp, codename, suffix

def timestamp():
    return time.time()

def dmesg(msg, lvl=1):
    if DEBUG: msg = '[DEBUG][%f] %r' % (timestamp(), msg)
    if VERBOSITY >= lvl:
        print(msg)

def check_output_basedir(loc):
    if not os.path.exists(loc):
        dmesg("Fatal: --outpath '%s' does not exist." % loc, 1)
        sys.exit()

    if os.path.isdir(loc):
        if os.access(loc, os.W_OK):
            return loc
        else:
            dmesg('Fatal: cannot write to %s (check permissions)' % loc, 1)
            sys.exit()
    else:
        dmesg('Fatal: %s exists but is not a directory.' % loc, 1)
        sys.exit()

def construct_filename(codename, args):
    if args['--json']:
        suffix = 'json'
    else:
        suffix = 'csv'

    outdir = os.path.join(check_output_basedir(args['--outpath']), args['--name'])
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    if args['--timestamp']:
        filename = TIMESTAMP_FNAME % (args['--timestamp'], codename, suffix)
    else:
        filename = DEFAULT_FNAME % (codename, suffix)
    return os.path.join(outdir, filename)

def write_data(output, codename, args):
    if args['--outpath']:
        outpath = construct_filename(codename, args)
        datafile = open(outpath, 'wb')
        try:
            datafile.write(output+'\n')
            dmesg('wrote %s' % outpath, 1)
        except Exception as e:
            dmesg('Error writing to file: %r' % e, 0)
        datafile.close()
    else:
        print(output)

def calculate_verbosity(args):
    global VERBOSITY
    if args['--quiet']:
        VERBOSITY=0
        return

    if args['--verbose'] or args['--debug']: 
        VERBOSITY += 1

def run_metrics_object(InteropObject, title, args):
    dmesg('%s: running' % title, 2)
    try:
        if args['--csv'] or args['--json']:
            dump(InteropObject, args)
        else:
            dmesg(title, 1)
            dmesg('-' * len(title), 1)
            dmesg('%s' % InteropObject(), 1)
    except InteropFileNotFoundError:
        dmesg('%s: File not found\n' % title, 1)

    dmesg('%s: finished' % title, 2)

def dump(InteropObject, args):
    try:
        metricobj = InteropObject()
        if args['--csv']:
            write_data(metricobj.to_csv(), metricobj.codename, args)
        elif args['--json']:
            write_data(metricobj.to_json(), metricobj.codename, args)

    except AttributeError:
        dmesg('Metadata has no CSV or JSON output.\n', 2)

def print_meta(metaobj, args):
    dmesg('Name:   %s' % args['--name'], 1)
    dmesg('Run ID: %s' % metaobj.runID, 1)
    dmesg('%s\n' % metaobj, 1)

def main():
    args = docopt(__doc__, version='0.5.7')

    if args['--interactive']:
        from IPython import embed
        myDataset = InteropDataset(args['<datapath>']) 
        embed()
        sys.exit()
    else:
        calculate_verbosity(args)

        ID = InteropDataset(args['<datapath>'])

        if args['--name']=='meta.runID':
            args['--name'] = ID.meta.runID

        if args['--all'] or args['--meta']:
            print_meta(ID.meta, args)
            
        if args['--timestamp']:
            args['--timestamp'] = timestamp()

        if args['--all'] or args['--tile']:
            run_metrics_object(ID.TileMetrics, "TILE METRICS", args)
        if args['--all'] or args['--quality']:
            run_metrics_object(ID.QualityMetrics, "QUALITY METRICS", args)
        if args['--all'] or args['--index']:
            run_metrics_object(ID.IndexMetrics, "INDEXING METRICS", args)
        if args['--all'] or args['--error']:
            run_metrics_object(ID.ErrorMetrics, "ERROR METRICS", args)
        if args['--all'] or args['--corint']:
            run_metrics_object(ID.CorrectedIntensityMetrics, "CORRECTED INTENSITY", args)
        if args['--all'] or args['--extraction']:
            run_metrics_object(ID.ExtractionMetrics, "EXTRACTION METRICS", args)
        if args['--all'] or args['--control']:
            run_metrics_object(ID.ControlMetrics, "CONTROL METRICS", args)

if __name__=='__main__':
    main()
